@@ -6,14 +6,13 @@ let () = Sqlite3.init
 let monot_encode s = Base64.encode ~linelength:72 s
 let monot_decode s = Base64.decode ~accept_spaces:true s
 
-let sql_escape s =
-  String.concat "''" (string_split ~collapse:false '\'' s)
-
 let may_decode base64 v =
   if base64 then monot_decode v else v
 
-let acc_one_col base64 acc row =
-  may_decode base64 row.(0) :: acc
+let blob_col base64 stmt n =
+  may_decode base64 (Sqlite3.column_blob stmt n)
+let acc_one_col base64 acc stmt =
+  blob_col base64 stmt 0 :: acc
 
 
 
@@ -21,7 +20,7 @@ let setup_sqlite ?busy_handler db =
   if Viz_misc.debug "sql"
   then
     Sqlite3.trace_set db
-      (fun s -> prerr_string "### sql: " ; prerr_endline s) ;
+      (fun s -> Printf.eprintf "### %2.3f sql: %s\n%!" (Sys.time ()) s) ;
   Sqlite3.exec db "PRAGMA temp_store = MEMORY" ;
   may
     (Sqlite3.busy_set db)
@@ -53,23 +52,24 @@ let uses_base64 rosters schema_id =
 let fetch_pubkeys db base64 tbl =
   Sqlite3.fetch db
     "SELECT id, keydata, ROWID FROM public_keys"
-    (fun () -> function 
-      | [| id; data; rowid |] -> 
-	  begin
-	    try 
-	      let data = may_decode base64 data in
-	      let key = Crypto.decode_rsa_pubkey data in
-	      Hashtbl.add tbl id (key, int_of_string rowid - 1)
-	    with Failure _ -> () 
-	  end
-      | _ -> ())
+    (fun () stmt ->
+      let id    = Sqlite3.column_text stmt 0 in
+      let data  = blob_col base64 stmt 1 in
+      let rowid = Sqlite3.column_int stmt 2 in
+      try 
+	let key = Crypto.decode_rsa_pubkey data in
+	Hashtbl.add tbl id (key, rowid - 1)
+      with Failure _ -> ())
     ()
 
 let fetch_branches base64 db =
   List.sort compare
     (Sqlite3.fetch db
-       "SELECT DISTINCT value FROM revision_certs WHERE name = 'branch'"
-       (acc_one_col base64)
+       "SELECT value, COUNT(id) FROM revision_certs WHERE name = 'branch' GROUP BY value"
+       (fun acc s ->
+	 let b = blob_col base64 s 0 in
+	 let n = Sqlite3.column_int s 1 in
+	 (b, n) :: acc)
        [])
 
 let view_name_domain = "ids_of_branch"
@@ -77,129 +77,145 @@ let view_name_limit  = "ids_of_branch_with_date_limit"
 
 let sql_of_bool b = if b then `INT 1 else `INT 0
 
-let id_set_add_if t v s =
-  if t && v <> "" then IdSet.add v s else s
-
-let add_node kind id rel_id rel nodes =
-  try
-    let current_node = NodeMap.find id nodes in
-    if List.mem_assoc rel_id current_node.family
-    then 
-      nodes
-    else 
-      NodeMap.add id 
-	{ current_node with family = (rel_id, rel) :: current_node.family }
-	nodes
-  with Not_found -> 
-    NodeMap.add id 
-      { id = id ; kind = kind ; family = [ rel_id, rel ] } 
-      nodes
 
 
-let process_ancestry_row g parent parent_kind child child_kind =
-  assert (parent_kind = REGULAR || child_kind = REGULAR) ;
-  assert (parent <> "" && child <> "") ; 
-  { nodes    = (add_node parent_kind parent child CHILD (
-		add_node child_kind child parent PARENT g.nodes)) ;
-
-    ancestry = 
-      EdgeMap.add (parent, child)
-        (if parent_kind = REGULAR && child_kind = REGULAR 
-         then SAME_BRANCH 
-         else BRANCHING)
-        g.ancestry ;
-
-    neighbour_nodes = (id_set_add_if (parent_kind <> REGULAR) parent (
-		       id_set_add_if (child_kind  <> REGULAR) child
-			 g.neighbour_nodes))
-  }
-
-let process_ancestry_row_simple g = function
-  | [| "" ; child |] when not (NodeMap.mem child g.nodes) -> 
-      let new_node = { id = child ; kind = REGULAR ; family = [] } in
-      let nodes = NodeMap.add child new_node g.nodes in
-      { g with nodes = nodes }
-
-  | [| parent ; child |] -> 
-      process_ancestry_row g 
-	parent REGULAR
-	child REGULAR
-
-  | _ -> g
-
-let process_ancestry_row_neigh_out g = function
-  | [| parent ; child ; "0" |] -> 
-	process_ancestry_row g 
-	  parent REGULAR
-	  child NEIGHBOUR_OUT
-  | [| parent ; child ; _ |] ->       
-      { g with ancestry = EdgeMap.add (parent, child) SAME_BRANCH g.ancestry }
-  | _ -> g
-
-let process_ancestry_row_neigh_in g = function
-  | [| parent ; child ; "0" |] -> 
-      process_ancestry_row g 
-	parent NEIGHBOUR_IN
-	child REGULAR
-  | [| parent ; child ; _ |] ->       
-      { g with ancestry = EdgeMap.add (parent, child) SAME_BRANCH g.ancestry }
-  | _ -> g
-
-
-let number_of_parent node =
+let get_relative n r =
   List.fold_left
-    (fun n -> function 
-      | (_, PARENT) -> n + 1
-      | _ -> n)
+    (fun acc -> function
+      | (id, rel) when r = rel -> id :: acc
+      | _  -> acc)
+    []
+    n.family
+
+let count_parents n = 
+  List.length (get_relative n PARENT)
+
+let find_merge_nodes g =
+  NodeMap.iter 
+    (fun id node ->
+      if node.kind = REGULAR && count_parents node > 1
+      then node.kind <- MERGE)
+    g.nodes
+
+let grab_one_int stmt id =
+  Sqlite3.bind_fetch stmt
+    [ `TEXT id ]
+    (fun _ stmt -> Sqlite3.column_int stmt 0)
     0
-    node.family
 
-let find_merge_nodes nodes =
-  NodeMap.fold
-    (fun id node m ->
-      if number_of_parent node > 1
-      then NodeMap.add id { node with kind = MERGE } m
-      else m)
-    nodes
-    nodes
+let count_all_parents db = 
+  let stmt =
+      Sqlite3.prepare_one
+      db "SELECT COUNT(parent) FROM revision_ancestry WHERE parent != '' AND child = ?" in
+  fun id ->
+    grab_one_int stmt id
 
-let process_branching_edge_row g = function
-  | [| parent; child |] ->
-      { g with ancestry = EdgeMap.add (parent, child) BRANCHING g.ancestry }
-  | _ -> g
+let count_regular_children db = 
+  let stmt =
+    Sqlite3.prepare_one db 
+      (Printf.sprintf 
+	 "SELECT COUNT(child) FROM revision_ancestry, %s WHERE parent = ? AND child = id"
+	 view_name_domain) in
+  fun id ->
+    grab_one_int stmt id
 
+let is_interesting_neighbour_out db = 
+  let count_p = count_all_parents db in
+  let count_c = count_regular_children db in
+  let start_of_branch id = 
+    count_p id = 1 in
+  let end_of_branch p_id =
+    count_c p_id = 0 in
+  fun id_parent id ->
+    start_of_branch id || end_of_branch id_parent
 
 let fetch_children db = 
-  let stmt = lazy 
-      (Sqlite3.prepare_one
-	 db "SELECT child FROM revision_ancestry WHERE parent = ?") in
+  let stmt = 
+    Sqlite3.prepare_one db 
+      "SELECT child FROM revision_ancestry WHERE parent = ?" in
   fun id f init ->
-    let stmt = Lazy.force stmt in
-    Sqlite3.reset stmt ;
-    Sqlite3.bind stmt 1 (`TEXT id) ;
-    Sqlite3.fold_rows
+    Sqlite3.bind_fetch 
+      stmt [ `TEXT id]
       (fun acc stmt -> f acc (Sqlite3.column_text stmt 0))
       init
-      stmt
+
+let collect_tags db base64 view g =
+  Sqlite3.fetch_f db
+    "SELECT C.id, C.value FROM revision_certs AS C, %s AS D WHERE name = 'tag' AND C.id = D.id"
+    view
+    (fun () stmt ->
+      let id = Sqlite3.column_text stmt 0 in
+      let n = NodeMap.find id g.nodes in
+      let tag = blob_col base64 stmt 1 in
+      n.kind <- TAGGED tag)
+    ()
 
 
-let collect_tags db nodes =
-  let stmt = Sqlite3.prepare_one db 
-      "SELECT value FROM revision_certs WHERE id = ? AND name = 'tag'" in
-  NodeMap.fold
-    (fun id node m ->
-      Sqlite3.reset stmt ;
-      Sqlite3.bind stmt 1 (`TEXT id) ;
-      Sqlite3.fold_rows
-	(fun m stmt -> 
-	  let v = Sqlite3.column_text stmt 0 in
-	  NodeMap.add id { node with kind = TAGGED v } m)
-	m
-	stmt)
-    nodes
-    nodes
+let ensure_node g id k =
+  try NodeMap.find id g.nodes, g
+  with Not_found ->
+    let n = { id = id ; kind = k ; family = [] } in
+    n, { g with nodes = NodeMap.add id n g.nodes }
 
-let fetch_agraph_with_view db (query, query_limit) =
+let connect_nodes n1 n2 =
+  n1.family <- (n2.id, CHILD)  :: n1.family ;
+  n2.family <- (n1.id, PARENT) :: n2.family
+
+let add_edge g id1 id2 ek =
+  { g with ancestry = EdgeMap.add (id1, id2) ek g.ancestry }
+
+let add_nodes_with_edge g id1 k1 id2 k2 ek = 
+  let n1, g = ensure_node g id1 k1 in
+  let n2, g = ensure_node g id2 k2 in
+  connect_nodes n1 n2 ;
+  add_edge g id1 id2 ek
+      
+
+let process_regular_node g s =
+  let id = Sqlite3.column_text s 0 in
+  let _, g = ensure_node g id REGULAR in
+  g
+
+let process_neighb_in g s =
+  let id    = Sqlite3.column_text s 0 in
+  let child = Sqlite3.column_text s 1 in
+  assert (NodeMap.mem child g.nodes) ;
+  add_nodes_with_edge g id NEIGHBOUR_IN child REGULAR BRANCHING_NEIGH
+
+let process_neighb_out all db =
+  let is_interesting = is_interesting_neighbour_out db in
+  fun g s ->
+    let parent = Sqlite3.column_text s 0 in
+    let id     = Sqlite3.column_text s 1 in
+    assert (NodeMap.mem parent g.nodes) ;
+    if all || is_interesting parent id
+    then add_nodes_with_edge g parent REGULAR id NEIGHBOUR_OUT BRANCHING_NEIGH
+    else g
+
+let process_ancestry g s =
+  let parent = Sqlite3.column_text s 0 in
+  let child  = Sqlite3.column_text s 1 in
+  assert (NodeMap.mem parent g.nodes) ;
+  assert (NodeMap.mem child  g.nodes) ;
+  assert (not (EdgeMap.mem (parent, child) g.ancestry)) ;
+  add_nodes_with_edge g parent REGULAR child REGULAR SAME_BRANCH
+
+let process_branching_edge g s =
+  let parent = Sqlite3.column_text s 0 in
+  let child  = Sqlite3.column_text s 1 in
+  let e = parent, child in
+  try
+    if EdgeMap.find e g.ancestry = SAME_BRANCH
+    then add_edge g parent child BRANCHING
+    else g
+  with Not_found ->
+    g
+      
+
+let fetch_agraph_with_view db base64 query =
+  let { dom = query ; 
+	lim = query_limit ; 
+	all_propagates = all_propagates } = query in
   let view_name_limit =
     if query_limit <> QUERY_NO_LIMIT
     then view_name_limit
@@ -207,31 +223,48 @@ let fetch_agraph_with_view db (query, query_limit) =
 
   let agraph = Viz_types.empty_agraph in
 
-  (* grab all node ids and edges we're interested in *)
+  (* grab all our main nodes *)
   let agraph =
-    Sqlite3.fetch_f db process_ancestry_row_simple agraph 
-      "SELECT parent, child FROM revision_ancestry, %s \
-        WHERE parent = id AND child IN %s" view_name_limit view_name_limit in
+    Sqlite3.fetch_f db
+      "SELECT id FROM %s" view_name_limit
+      process_regular_node agraph in
 
-  (* also grab neighbor nodes *)
+  (* neighbor IN *)
   let agraph =
-    Sqlite3.fetch_f db process_ancestry_row_neigh_out agraph 
-      "SELECT parent, child, child IN %s FROM revision_ancestry 
-        WHERE parent IN %s AND child NOT IN %s"
-      view_name_domain view_name_limit view_name_limit in
+    Sqlite3.fetch_f db 
+        "SELECT parent, child \
+           FROM %s AS D1 \
+           JOIN revision_ancestry ON D1.id = child \
+LEFT OUTER JOIN %s AS D2 ON D2.id = parent \
+          WHERE D2.id ISNULL AND parent != ''"
+      view_name_limit view_name_domain
+      process_neighb_in agraph in
+
+  (* neighbor OUT *)
   let agraph =
-    Sqlite3.fetch_f db process_ancestry_row_neigh_in agraph 
-      "SELECT parent, child, parent IN %s FROM revision_ancestry 
-        WHERE child IN %s AND parent != '' AND parent NOT IN %s"
-      view_name_domain view_name_limit view_name_limit in
+    Sqlite3.fetch_f db
+        "SELECT parent, child \
+           FROM %s AS D1 \
+           JOIN revision_ancestry ON D1.id = parent \
+LEFT OUTER JOIN %s AS D2 ON D2.id = child \
+          WHERE D2.id ISNULL"
+      view_name_limit view_name_domain
+      (process_neighb_out all_propagates db) agraph in
+
+  (* ancestry *)
+  let agraph =
+    Sqlite3.fetch_f db
+      "SELECT parent, child \
+         FROM %s AS D1, revision_ancestry, %s AS D2 \
+        WHERE D1.id = parent AND child = D2.id"
+      view_name_limit view_name_limit
+      process_ancestry agraph in
 
   (* find merge/propagate nodes (they have more than one parent) *)
-  let agraph =
-    { agraph with nodes = find_merge_nodes agraph.nodes } in
+  find_merge_nodes agraph ;
 
   (* get tags *)
-  let agraph =
-    { agraph with nodes = collect_tags db agraph.nodes } in
+  collect_tags db base64 view_name_limit agraph ;
 
   (* determine the branching edges *)
   let agraph =
@@ -243,16 +276,17 @@ let fetch_agraph_with_view db (query, query_limit) =
       | _ -> 
 	  (* we need another database query *)
 	  Sqlite3.fetch_f db
-	    process_branching_edge_row agraph
             "SELECT parent, child \
-               FROM revision_ancestry AS A \
-              WHERE A.child IN %s AND A.parent != '' AND \
-	            NOT EXISTS \
+               FROM %s, revision_ancestry \
+              WHERE id = child \
+                AND parent != '' \
+                AND NOT EXISTS \
                      (SELECT P.id FROM revision_certs AS C, revision_certs AS P \
-                       WHERE C.id = A.child AND P.id = A.parent \
+                       WHERE C.id = child AND P.id = parent \
                              AND C.name = 'branch' AND P.name = 'branch' \
 			     AND C.value = P.value)"
             view_name_limit
+	    process_branching_edge agraph
     end in
 
   (* reconnect disconnected components *)
@@ -272,17 +306,14 @@ let encode_quote base64 s =
     String.concat "" [ "X\'" ; Viz_misc.hex_enc s ; "\'" ]
 
 let fetch_with_view query base64 db f =
-  let (query_domain, query_limit) = query in
+  let { dom = query_domain ; lim = query_limit } = query in
 
   let register_date_p () =
     match query_limit with
     | QUERY_BETWEEN (d1, d2) ->
 	Sqlite3.create_fun_1 db "date_p"
 	  (fun arg ->
-	    let d = 
-	      if base64
-	      then monot_decode (Sqlite3.value_text arg)
-	      else Sqlite3.value_blob arg in
+	    let d = may_decode base64 (Sqlite3.value_text arg) in
 	    sql_of_bool (d1 <= d && d <= d2))
     | _ -> () in
 
@@ -290,7 +321,7 @@ let fetch_with_view query base64 db f =
     match query_domain with
     | QUERY_ALL -> Printf.sprintf
 	  "CREATE TEMP TABLE %s AS \
-           SELECT DISTINCT id FROM revision_certs
+           SELECT DISTINCT id FROM revision_certs \
             WHERE name = 'branch'" view_name_domain
     | QUERY_BRANCHES q -> Printf.sprintf
 	  "CREATE TEMP TABLE %s AS \
@@ -303,8 +334,8 @@ let fetch_with_view query base64 db f =
   let view_query_date_limit () =
     Printf.sprintf
       "CREATE TEMP TABLE %s AS \
-       SELECT DISTINCT id FROM revision_certs \
-        WHERE name = 'date' AND id IN %s AND date_p(value)"
+       SELECT DISTINCT id FROM %s NATURAL JOIN revision_certs \
+        WHERE name = 'date' AND date_p(value)"
       view_name_limit view_name_domain in
   
   Viz_misc.bracket
@@ -321,7 +352,7 @@ let fetch_with_view query base64 db f =
 	Sqlite3.exec_f db 
 	  "CREATE INDEX %s__id ON %s (id)" view_name_limit view_name_limit
       end)
-    ~action:(fun () -> f db query)
+    ~action:(fun () -> f db base64 query)
     ~after:(fun () -> 
       if query_limit <> QUERY_NO_LIMIT
       then begin 
@@ -352,8 +383,9 @@ let fetch_revision_set rostered b64 db id =
   decode_and_parse_revision 
     rostered
     (List.hd
-       (Sqlite3.fetch_f db (acc_one_col b64) []
-	  "SELECT data FROM revisions WHERE id = '%s'" id))
+       (Sqlite3.fetch_v db 
+	  "SELECT data FROM revisions WHERE id = ?" [`TEXT id]
+	  (acc_one_col b64) []))
 
 let verify_cert_sig pubkeys keypair name id v signature =
   try
@@ -365,34 +397,34 @@ let verify_cert_sig pubkeys keypair name id v signature =
     else SIG_BAD
   with Not_found -> SIG_UNKNOWN
 
-let process_certs pubkeys b64 acc = function
-  | [| id; name; v; keypair; signature |] ->
-      let dec_v = may_decode b64 v in
-      let dec_sig = may_decode b64 signature in
-      { c_id = id ; 
-	c_name = name ; 
-	c_value = dec_v ;
-	c_signer_id = keypair ; 
-	c_signature = verify_cert_sig pubkeys keypair name id dec_v dec_sig } :: acc
-  | _ -> acc
+let process_certs pubkeys b64 acc s =
+    let id      = Sqlite3.column_text s 0 in
+    let name    = Sqlite3.column_text s 1 in
+    let dec_v   = blob_col b64 s 2 in
+    let keypair = Sqlite3.column_text s 3 in
+    let dec_sig = blob_col b64 s 4 in
+    { c_id = id ; 
+      c_name = name ; 
+      c_value = dec_v ;
+      c_signer_id = keypair ; 
+      c_signature = verify_cert_sig pubkeys keypair name id dec_v dec_sig } :: acc
 
 let fetch_certs db pubkeys b64 id =
-  Sqlite3.fetch_f db (process_certs pubkeys b64) []
+  Sqlite3.fetch_v db
     "SELECT id, name, value, keypair, signature \
-       FROM revision_certs WHERE id = '%s'" id
-
+       FROM revision_certs WHERE id = ?" [`TEXT id]
+    (process_certs pubkeys b64) []
 let prepare_fetch_one_cert_signer db =
   Sqlite3.prepare_one db
     "SELECT keypair FROM revision_certs WHERE id = ? AND name = ?"
 let prepare_fetch_one_cert_value db =
   Sqlite3.prepare_one db
-    "SELECT value   FROM revision_certs WHERE id = ? AND name = ?"
+    "SELECT value FROM revision_certs WHERE id = ? AND name = ?"
 
 let fetch_one_cert_field stmt id name kind =
-  Sqlite3.reset stmt ;
-  Sqlite3.bind stmt 1 (`TEXT id) ;
-  Sqlite3.bind stmt 2 (`TEXT name) ;
-  Sqlite3.fold_rows 
+  Sqlite3.bind_fetch
+    stmt
+    [ `TEXT id ; `TEXT name ]
     (fun acc stmt ->
       let v = Sqlite3.column_text stmt 0 in
       match kind with
@@ -400,20 +432,21 @@ let fetch_one_cert_field stmt id name kind =
       | `VALUE -> v :: acc
       | `VALUE_B64 -> monot_decode v :: acc)
     []
-    stmt
 
 let get_matching_cert db b64 name p =
   List.rev
-    (Sqlite3.fetch_f db
-       (fun acc -> function
-	 | [| id; v |] -> 
-	     let dv = may_decode b64 v in
-	     if p dv 
-	     then (id, dv) :: acc
-	     else acc
-	 | _ -> acc)
-       []
-       "SELECT id, value FROM revision_certs WHERE name = '%s'" name)
+    (Sqlite3.fetch_v db
+       "SELECT id, value FROM revision_certs WHERE name = ?"
+       [`TEXT name]
+       (fun acc s ->
+	 let v  = blob_col b64 s 1 in
+	 if p v 
+	 then begin 
+	   let id = Sqlite3.column_text s 0 in
+	   (id, v) :: acc
+	 end
+	 else acc)
+       [])
 
 
 
@@ -461,8 +494,11 @@ type t = {
 
 let sqlite_try f db =
   try f db.db
-  with Sqlite3.Error (_, msg) ->
-    Viz_types.errorf "Error processing database %s:\n%s" db.filename msg
+  with 
+  | Sqlite3.Error ((Sqlite3.LOCKED | Sqlite3.BUSY), _) as exn ->
+      raise exn
+  | Sqlite3.Error (_, msg) ->
+      Viz_types.errorf "Error processing database %s:\n%s" db.filename msg
 
 
 
