@@ -11,7 +11,7 @@ val with_progress : (unit -> unit) -> (t -> 'a) -> t -> 'a
 
 val get_filename : t -> string
 
-val fetch_branches : t -> string list
+val fetch_branches : t -> (string * int) list
 val fetch_ancestry_graph : t -> query -> agraph
 val fetch_revision           : t -> string -> node_data
 val fetch_certs_and_revision : t -> string -> node_data
