@@ -364,11 +364,11 @@ let ui_info = "\
       <menu action='FileMenu'>\
         <menuitem action='Open'/>\
         <menuitem action='Close'/>\
+        <menuitem action='New'/>\
         <menuitem action='Prefs'/>\
         <menuitem action='Quit'/>\
       </menu>\
       <menu action='ViewMenu'>\
-        <menuitem action='New'/>\
         <menuitem action='Refresh'/>\
         <menuitem action='Zoom_in'/>\
         <menuitem action='Zoom_out'/>\
@@ -408,7 +408,7 @@ let make_groups () =
   let g_db = GAction.action_group ~name:"db" () in
   GAction.add_actions g_db [
   add "Close" ~stock:`CLOSE ~tooltip:"Close the database" ;
-  add "New"   ~stock:`NEW   ~tooltip:"View a monotone ancestry graph" ] ;
+  add "New"   ~stock:`NEW   ~label:"New view" ~tooltip:"View a monotone ancestry graph" ] ;
   g_db#set_sensitive false ;
   let g_popup = GAction.action_group ~name:"popup" () in
   GAction.add_actions g_popup [
@@ -569,7 +569,7 @@ let setup ({ manager = ui } as m) ctrl =
   action_connect "/menubar/ViewMenu/Refresh"  ctrl#reload ;
   action_connect "/menubar/FileMenu/Prefs"    ctrl#show_prefs ;
   action_connect "/menubar/ViewMenu/Query"    ctrl#show_search ;
-  action_connect "/menubar/ViewMenu/New"      ctrl#show_view ;
+  action_connect "/menubar/FileMenu/New"      ctrl#show_view ;
   action_connect "/popup/Certs"      
     (fun () -> ctrl#display_certs (get_popup_data m).popup_id) ;
   action_connect "/popup_cert/Copy_cert"
